
import json

from .Form803Elements.FormElement import FormElement, FormParams, FormProps, FormCommands, calc_offset
from .Form8x import Form8x
from ... import helper
from ...ext_exception import ExtException

OLD_FORM = '0'
UPR_FORM = '1'


class Form803(Form8x):
    version = '803'

    def __init__(self, *, obj_name=None, options=None):
        super().__init__(obj_name=obj_name, options=options)
        self.command_panels = []
        self.params = []
        self.commands = []

    def decode_data(self, src_dir, uuid):
        _header_obj = self.get_decode_obj_header(self.header['data'])
        self.header['Включать в содержание справки'] = _header_obj[1][2]
        self.header['Тип формы'] = _header_obj[1][3]
        self.header['Версия803'] = _header_obj[1][0]

        try:
            self.header['Расширенное представление'] = _header_obj[2]
        except IndexError:
            pass
        try:
            self.header['ХЗ1'] = _header_obj[1][4]
        except IndexError:
            pass

        if self.header['Тип формы'] != OLD_FORM:
            self.decode_form0(src_dir, uuid)

    def decode_includes(self, src_dir, dest_dir, dest_path, header_data):
        if self.header.get('Тип формы') == OLD_FORM:
            self.decode_old_elements()
            return
        if not self.form or not self.form[0]:
            return
        try:
            supported_form = ['4-49', '3-49']  # not supported 27-18, контур вероятно обычные формы
            current_form = f'{self.form[0][0][0]}-{self.form[0][0][1][0]}'
            if current_form not in supported_form:
                return
        except:
            return

        backup = json.dumps(self.form)
        try:
            self.decode_elements(src_dir, dest_dir, dest_path, header_data)
            self.params = FormParams.decode_list(self, self.form[0][0])
            self.commands = FormCommands.decode_list(self, self.form[0][0])
            self.props = FormProps.decode_list(self, self.form[0][0])
        except Exception as err:
            self.form = json.loads(backup)
            pass  # todo если какие то елементы формы не разбираются, не прерываем
            # raise ExtException(parent=err, message='Ошибка при разборе формы')

    def decode_elements(self, src_dir, dest_dir, dest_path, header_data):
        index = self.get_form_elem_index()
        root_data = self.form[0][0][1]

        # index_panel_count = index[1]
        # form_panels_count = int(root_data[index_panel_count])
        # if form_panels_count:
        #     self.command_panels = [root_data[index_panel_count + 1]]
        #     root_data[index_panel_count] = 'В отдельном файле'
        #     del root_data[index_panel_count + 1]

        index_root_element_count = index[0]
        form_items_count = int(root_data[index_root_element_count])
        if form_items_count:
            self.elements = FormElement.decode_list(self, root_data, index_root_element_count)
            self.elements_data = dict(sorted(self.elements_data.items()))
        pass

    def get_form_elem_index(self):
        try:
            root_data = self.form[0][0][1]
            index_command_panel_count = calc_offset([(18, 2), (3, 0)], root_data)
            command_panel_count = int(root_data[index_command_panel_count])
            index_root_elem_count = index_command_panel_count + command_panel_count + 1
            return index_root_elem_count, index_command_panel_count
        except Exception as err:
            raise ExtException(
                message='случай требующий анализа, предоставьте образец формы разработчикам',
                detail=f'{self.header["name"]}, {err}')

    def write_decode_object(self, dest_dir, dest_path, file_name):
        super().write_decode_object(dest_dir, dest_path, file_name)
        if self.commands:
            helper.json_write(self.commands, self.new_dest_dir, f'{file_name}.commands{self.version}.json')
        if self.params:
            helper.json_write(self.params, self.new_dest_dir, f'{file_name}.params{self.version}.json')
        if self.command_panels:
            helper.json_write(self.command_panels, self.new_dest_dir, f'{file_name}.panels{self.version}.json')

    def decode_form1(self, src_dir, uuid):
        try:
            form = helper.brace_file_read(src_dir, f'{uuid}.1')
        except FileNotFoundError:
            return
        self.form.append(form)

    def decode_code(self, src_dir):
        if self.header['Тип формы'] == OLD_FORM:
            self.decode_old_form(src_dir)
        else:
            super().decode_code(src_dir)
        self.decode_form1(src_dir, self.header['uuid'])

    # def encode_header(self):
    #     return [[
    #         "1",
    #         [
    #             "1",
    #             [
    #                 "0",
    #                 self.encode_header_title(),
    #                 self.header.get('Расширенное представление', ['0'])
    #             ]
    #         ],
    #         "0"
    #     ]]

    def encode_header_title(self):
        version = self.header.get('Версия803', '13')
        header_title = [
            version,
            [
                self.header['h0'],
                [
                    self.header['h1_0'],
                    "0",
                    self.header['uuid']
                ],
                helper.str_encode(self.header['name']),
                helper.encode_name2(self.header),
                helper.str_encode(self.header['comment']),
                *self.header['h5'],
            ],
            self.header['Включать в содержание справки'],
            self.header['Тип формы']
        ]
        try:
            header_title.append(self.header['ХЗ1'])
        except KeyError:
            pass
        return header_title

    def encode_data(self):
        if self.header['Тип формы'] == OLD_FORM:
            return
        try:
            _code = self.code.pop('obj', "")
            _code = self.quotes.sub(r'""', _code)
            form0 = self.form[0]
            self.getset_form_code(form0, helper.str_encode(_code), self.header)
        except IndexError:
            pass
        return self.form

    def encode_nested_includes(self, src_dir, file_name, dest_dir, parent_id):
        if self.header.get('Тип формы') == OLD_FORM:
            self.encode_old_elements(src_dir, file_name, dest_dir, parent_id)
            return

        if not self.form or not self.form[0]:
            return
        try:
            supported_form = ['4-49', '3-49']
            if f'{self.form[0][0][0]}-{self.form[0][0][1][0]}' not in supported_form:
                return
        except:
            return
        try:
            self.encode_elements(src_dir, file_name, dest_dir, self.version)
            FormParams.encode_list(self, src_dir, file_name, self.version)
            FormCommands.encode_list(self, src_dir, file_name, self.version)
            FormProps.encode_list(self, src_dir, file_name, self.version)
        except Exception as err:
            raise ExtException(parent=err, message='Ошибка при разборе формы')

    def encode_elements(self, src_dir, file_name, dest_dir, version):
        index = self.get_form_elem_index()
        root_data = self.form[0][0][1]

        # index_panel_count = index[1]
        # form_panels_count = int(root_data[index_panel_count])
        # if form_panels_count == 'В отдельном файле':
        #     self.command_panels = helper.json_read(src_dir, f'{file_name}.panels{version}.json')

        index_root_element_count = index[0]
        if root_data[index_root_element_count] == 'Дочерние элементы отдельно':
            self.elements = helper.json_read(src_dir, f'{file_name}.elements.tree{version}.json')
            self.elements_data = helper.json_read(src_dir, f'{file_name}.elements.data{version}.json')
            # root_data[index_root_element_count] = str(len(self.elements))
            FormElement.encode_list(self, self.elements, root_data, index_root_element_count)

    def write_encode_object(self, dest_dir):
        if self.header['Тип формы'] == OLD_FORM:
            self.write_old_encode_object(dest_dir)
        else:
            helper.brace_file_write(self.encode_header(), dest_dir, self.header["uuid"])
            self.file_list.append(self.header["uuid"])
            if self.form:
                file_name = f'{self.header["uuid"]}.0'
                helper.brace_file_write(self.form[0], dest_dir, file_name)
                self.file_list.append(file_name)
        if self.form and len(self.form) > 1:
            file_name = f'{self.header["uuid"]}.1'
            helper.brace_file_write(self.form[1], dest_dir, file_name)
            self.file_list.append(file_name)

    def encode_empty_form(self):
        return [[
            "4",
            [
                "49",
                "0",
                "0",
                "0",
                "0",
                "1",
                "0",
                "0",
                "00000000-0000-0000-0000-000000000000",
                "1",
                [
                    "1",
                    "0"
                ],
                "0",
                "0",
                "1",
                "1",
                "1",
                "0",
                "1",
                "0",
                [
                    "0",
                    "1",
                    "0"
                ],
                [
                    "0"
                ],
                "1",
                [
                    "22",
                    [
                        "-1",
                        "02023637-7868-4a5f-8576-835a76e0c9ba"
                    ],
                    "0",
                    "0",
                    "0",
                    "9",
                    "\"\"",
                    [
                        "1",
                        "0"
                    ],
                    [
                        "1",
                        "0"
                    ],
                    "0",
                    "1",
                    "0",
                    "0",
                    "0",
                    "2",
                    "2",
                    [
                        "3",
                        "4",
                        [
                            "0"
                        ]
                    ],
                    [
                        "7",
                        "3",
                        "0",
                        "1",
                        "100"
                    ],
                    [
                        "0",
                        "0",
                        "0"
                    ],
                    "1",
                    [
                        "0",
                        "0",
                        "1"
                    ],
                    "0",
                    "1",
                    "0",
                    "0",
                    "0",
                    "3",
                    "3",
                    "0"
                ],
                "0",
                "\"\"",
                "\"\"",
                "1",
                [
                    "22",
                    [
                        "0"
                    ],
                    "0",
                    "0",
                    "0",
                    "7",
                    "\"Navigator\"",
                    [
                        "1",
                        "0"
                    ],
                    [
                        "1",
                        "0"
                    ],
                    "0",
                    "1",
                    "0",
                    "0",
                    "0",
                    "2",
                    "2",
                    [
                        "3",
                        "4",
                        [
                            "0"
                        ]
                    ],
                    [
                        "7",
                        "3",
                        "0",
                        "1",
                        "100"
                    ],
                    [
                        "0",
                        "0",
                        "0"
                    ],
                    "0",
                    "0",
                    "1",
                    "0",
                    "1",
                    [
                        "11",
                        [
                            "0"
                        ],
                        "0",
                        "0",
                        "0",
                        "0",
                        "\"NavigatorExtendedTooltip\"",
                        [
                            "1",
                            "0"
                        ],
                        [
                            "1",
                            "0"
                        ],
                        "1",
                        "0",
                        "0",
                        "2",
                        "2",
                        [
                            "3",
                            "4",
                            [
                                "0"
                            ]
                        ],
                        [
                            "7",
                            "3",
                            "0",
                            "1",
                            "100"
                        ],
                        [
                            "0",
                            "0",
                            "0"
                        ],
                        "1",
                        [
                            "5",
                            "0",
                            "0",
                            "3",
                            "0",
                            [
                                "0",
                                "1",
                                "0"
                            ],
                            [
                                "3",
                                "4",
                                [
                                    "0"
                                ]
                            ],
                            [
                                "3",
                                "4",
                                [
                                    "0"
                                ]
                            ],
                            [
                                "3",
                                "0",
                                [
                                    "0"
                                ],
                                "0",
                                "1",
                                "0",
                                "48312c09-257f-4b29-b280-284dd89efc1e"
                            ]
                        ],
                        "0",
                        "1",
                        "2",
                        [
                            "1",
                            [
                                "1",
                                "0"
                            ],
                            "0"
                        ],
                        "0",
                        "0",
                        "1",
                        "0",
                        "0",
                        "1",
                        "0",
                        "3",
                        "3",
                        "0"
                    ],
                    "0",
                    "3",
                    "3",
                    "0"
                ],
                "1",
                "\"\"",
                "0",
                "0",
                "0",
                "0",
                "0",
                "0",
                "3",
                "3",
                "0",
                "0",
                "0",
                "100",
                "1",
                "1",
                "0",
                "0",
                "0",
                [
                    "49",
                    "0"
                ]
            ],
            "Код в отдельном файле",
            [
                "4",
                "1",
                [
                    "9",
                    [
                        "1"
                    ],
                    "0",
                    "\"Объект\"",
                    [
                        "1",
                        "0"
                    ],
                    [
                        "\"Pattern\"",
                        [
                            "\"#\"",
                            "7cadabb8-6ade-4026-bc22-5e7e789b0b4b"
                        ]
                    ],
                    [
                        "0",
                        [
                            "0",
                            [
                                "\"B\"",
                                "1"
                            ],
                            "0"
                        ]
                    ],
                    [
                        "0",
                        [
                            "0",
                            [
                                "\"B\"",
                                "1"
                            ],
                            "0"
                        ]
                    ],
                    [
                        "0",
                        "0"
                    ],
                    [
                        "0",
                        "0"
                    ],
                    "1",
                    "0",
                    "0",
                    "0",
                    [
                        "0",
                        "0"
                    ],
                    [
                        "0",
                        "0"
                    ]
                ],
                "0",
                "0",
                [
                    "#base64:77u/PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0iVVRGLTgiPz4NCjxTZXR0aW5ncyB4bWxucz0iaHR0cDovL3Y4LjFjLnJ1LzguMS9kYXRhLWNvbXBvc2l0aW9uLXN5c3RlbS9zZXR0aW5ncyIgeG1sbnM6ZGNzY29yPSJodHRwOi8vdjguMWMucnUvOC4xL2RhdGEtY29tcG9zaXRpb24tc3lzdGVtL2NvcmUiIHhtbG5zOnN0eWxlPSJodHRwOi8vdjguMWMucnUvOC4xL2RhdGEvdWkvc3R5bGUiIHhtbG5zOnN5cz0iaHR0cDovL3Y4LjFjLnJ1LzguMS9kYXRhL3VpL2ZvbnRzL3N5c3RlbSIgeG1sbnM6djg9Imh0dHA6Ly92OC4xYy5ydS84LjEvZGF0YS9jb3JlIiB4bWxuczp2OHVpPSJodHRwOi8vdjguMWMucnUvOC4xL2RhdGEvdWkiIHhtbG5zOndlYj0iaHR0cDovL3Y4LjFjLnJ1LzguMS9kYXRhL3VpL2NvbG9ycy93ZWIiIHhtbG5zOndpbj0iaHR0cDovL3Y4LjFjLnJ1LzguMS9kYXRhL3VpL2NvbG9ycy93aW5kb3dzIiB4bWxuczp4cz0iaHR0cDovL3d3dy53My5vcmcvMjAwMS9YTUxTY2hlbWEiIHhtbG5zOnhzaT0iaHR0cDovL3d3dy53My5vcmcvMjAwMS9YTUxTY2hlbWEtaW5zdGFuY2UiPg0KCTxvdXRwdXRQYXJhbWV0ZXJzLz4NCjwvU2V0dGluZ3M+"
                ]
            ],
            [
                "0",
                "0"
            ],
            [
                "0",
                "0"
            ],
            [
                "0",
                "0"
            ],
            [
                "0",
                "0"
            ],
            "0",
            "0"
        ]]



from ...MetaDataObject import MetaDataObject


class IncludeEmpty(MetaDataObject):

    @classmethod
    def decode(cls, src_dir, file_name, dest_dir, dest_path, version, parent_type=None):
        raise Exception('Так быть не должно, этот класс обслуживает вложенные объекты')

    @classmethod
    def decode_local_include(cls, parent, header_data, src_dir, dest_dir, dest_path, version):
        return

    def encode_object(self, src_dir, file_name, dest_dir):
        raise Exception('Так быть не должно, этот класс обслуживает вложенные объекты')



import copy
import json
import os
import shutil
import time
import uuid
from codecs import BOM_UTF8, BOM_UTF16_BE, BOM_UTF16_LE, BOM_UTF32_BE, BOM_UTF32_LE
from multiprocessing import Pool, cpu_count

from tqdm.auto import tqdm

from .ext_exception import ExtException
from .json_container_decoder import JsonContainerDecoder, BigBase64


def brace_file_read(path, file_name):
    _path = os.path.join(path, file_name)
    try:
        for code_page in ['utf-8-sig', 'windows-1251']:
            try:
                with open(_path, 'r', encoding=code_page) as file:
                    decoder = JsonContainerDecoder(src_dir=path, file_name=file_name)
                    data = decoder.decode_file(file)
                    return data
            except UnicodeDecodeError:
                continue
        raise ExtException(message=f'Unknown code page in file {file_name}')
    except (BigBase64, FileNotFoundError) as err:
        raise err from err
    except Exception as err:
        raise ExtException(parent=err, message='Ошибка чтения', detail=f'{err} в файле ({_path})')


def brace_file_write(data, path, file_name):
    _path = os.path.join(path, file_name)
    makedirs(path, exist_ok=True)
    try:
        with open(_path, 'w', encoding='utf-8') as file:
            decoder = JsonContainerDecoder(src_dir=path, file_name=file_name)
            raw_data = decoder.encode_root_object(data)
            decoder.write_data(path, file_name, raw_data)
    except Exception as err:
        raise ExtException(message='Ошибка записи', detail=f'{err} в файле ({_path})')


def json_read(path, file_name):
    _path = os.path.join(path, file_name)
    try:
        with open(_path, 'r', encoding='utf-8') as file:
            return json.load(file)
    except FileNotFoundError as err:
        raise err from err
    except Exception as err:
        raise ExtException(message='Ошибка чтения', detail=f'{err} в файле ({_path})')


def json_write(data, path, file_name):
    _path = os.path.join(path, file_name)
    makedirs(path, exist_ok=True)
    try:
        with open(_path, 'w', encoding='utf-8') as file:
            json.dump(data, file, ensure_ascii=False, indent=2)
    except Exception as err:
        raise ExtException(message='Ошибка записи', detail=f'{err} в файле ({_path})')


def txt_read(path, file_name, encoding='utf-8-sig'):
    try:
        return txt_read_detect_encoding(path, file_name, encoding=encoding)[0]
    except FileNotFoundError as err:
        raise err from err
    except Exception as err:
        raise ExtException(message='Ошибка чтения', detail=f'{err} в файле ({file_name})')


def txt_read_detect_encoding(path, file_name, encoding=None):
    _path = os.path.join(path, file_name)
    if encoding is None:
        encoding = detect_by_bom(_path, 'utf-8')
    with open(_path, 'r', encoding=encoding) as file:
        return file.read(), encoding


def txt_write(data, path, file_name, encoding='utf-8'):
    try:
        if data is None:
            return
        _path = os.path.join(path, file_name)
        makedirs(path, exist_ok=True)
        for i in range(3):
            try:
                with open(_path, 'w', encoding=encoding) as file:
                    file.write(data)
                return
            except PermissionError:
                time.sleep(0.5)
        raise PermissionError(_path)
    except Exception as err:
        raise ExtException(message='Ошибка записи файла', detail=f'{err} в файле {path}')


def bin_write(data, path, file_name):
    _path = os.path.join(path, file_name)
    makedirs(path, exist_ok=True)
    with open(_path, 'wb') as file:
        file.write(data)


def bin_read(path, file_name):
    _path = os.path.join(path, file_name)
    with open(_path, 'rb') as file:
        return file.read()


def decode_header(obj: dict, header: list):
    try:
        obj['uuid'] = header[1][2]
        uuid.UUID(obj['uuid'])

    except (ValueError, IndexError):
        raise ValueError('Заголовок определен не верно')

    obj['name'] = str_decode(header[2])
    obj['name2'] = {}
    count_locale = int(header[3][0])
    for i in range(count_locale):
        obj['name2'][str_decode(header[3][i * 2 + 1])] = str_decode(header[3][i * 2 + 2])
    comment = str_decode(header[4]).split(';')[0]  # удаляем имя файла и номер версии которую добавляем при сборке
    header[4] = str_encode(comment)
    obj['comment'] = comment
    obj['h1_0'] = header[1][0]
    obj['h0'] = header[0]
    obj['h5'] = header[5:]


def clear_dir(path: str) -> None:
    shutil.rmtree(path, ignore_errors=True)
    makedirs(path, exist_ok=True)


def str_encode(data: str) -> str:
    return f'"{data}"'


def encode_name2(header: dict):
    result = [str(len(header['name2']))]
    for elem in header['name2']:
        result.append(str_encode(elem))
        result.append(str_encode(header['name2'][elem]))
    return result


def str_decode(data: str) -> str:
    return data[1:-1]


def get_pool(*, pool: Pool = None, processes=None) -> Pool:
    if pool is not None:
        return pool
    if processes is None:
        processes = max(cpu_count() - 2, 1)  # чтобы система совсем не висла
    return Pool(processes)


def close_pool(local_pool: Pool, pool: Pool = None) -> None:
    if pool is None:
        local_pool.close()
        local_pool.join()


def run_in_pool(method, list_args, pool=None, title=None, need_result=False):
    _pool = get_pool(pool=pool)
    result = []
    try:
        with tqdm(desc=title, total=len(list_args)) as pbar:
            for _res in _pool.imap_unordered(method, list_args, chunksize=1):
                if need_result and _res:
                    result.extend(_res)
                pbar.update()
    except ExtException as err:
        raise ExtException(
            parent=err,
            action=f'run_in_pool {method.__qualname__}') from err
    finally:
        close_pool(_pool, pool)
    return result


def file_size(file):
    """
    Возвращает размер file-like объекта

    :param file: объекта файла
    :type file: BufferedReader
    :return: размер в байтах
    :rtype: int
    """
    pos = file.tell()
    file.seek(0, os.SEEK_END)
    size = file.tell()
    file.seek(pos)
    return size


def run_in_pool_encode_include(method, list_args, pool=None, title=None):
    _pool = get_pool(pool=pool)
    file_list = []
    include_index = {}
    object_task = []
    child_tasks = []
    try:
        with tqdm(desc=title, total=len(list_args)) as pbar:
            for _object_task, _child_tasks in _pool.imap_unordered(method, list_args, chunksize=1):
                if _child_tasks:
                    child_tasks.extend(_child_tasks)
                if isinstance(_object_task, list):
                    object_task.append(_object_task)
                elif isinstance(_object_task, dict):
                    parent_id = _object_task['parent_id']
                    obj_uuid = _object_task['obj_uuid']
                    obj_type = _object_task['obj_type']

                    if _object_task['file_list']:
                        file_list.extend(_object_task['file_list'])
                    if parent_id not in include_index:
                        include_index[parent_id] = {}
                    if obj_type not in include_index[parent_id]:
                        include_index[parent_id][obj_type] = []
                    include_index[parent_id][obj_type].append(obj_uuid)
                else:
                    raise NotImplementedError()
                pbar.update()
    except ExtException as err:
        raise ExtException(
            parent=err,
            action=f'run_in_pool {method.__qualname__}') from err
    # except Exception as err:
    #     raise ExtException(parent=err, detail=f'{method.__qualname__ {err.message}' action=f'run_in_pool {method.__qualname__}') from err
    finally:
        close_pool(_pool, pool)
    return file_list, include_index, object_task, child_tasks


def list_merge(*args):
    result = []
    for lst in args:
        if lst:
            result.extend(lst)
    return result


def get_class_metadata_object(name):
    return get_class(f'v8unpack.MetaDataObject.{name}.{name}')


def get_class(kls):
    try:
        parts = kls.split('.')
        module = ".".join(parts[:-1])
        m = __import__(module)
        for comp in parts[1:]:
            m = getattr(m, comp)
        return m
    except ImportError as e:
        # ошибки в классе  или нет файла
        raise ImportError(f'get_class({kls}: {str(e)}')
    except AttributeError as e:
        # Нет такого класса
        raise AttributeError(f'get_class({kls}: {str(e)}')
    except Exception as e:
        # ошибки в классе
        raise Exception(f'get_class({kls}: {str(e)}')


def detect_by_bom(path, default=None):
    boms = (
        ('utf-8-sig', BOM_UTF8),
        ('utf-32', BOM_UTF32_LE),
        ('utf-32', BOM_UTF32_BE),
        ('utf-16', BOM_UTF16_LE),
        ('utf-16', BOM_UTF16_BE),
    )

    with open(path, 'rb') as f:
        raw = f.read(4)  # will read less if the file is smaller
    for enc, bom in boms:
        if raw.startswith(bom):
            return enc
    return default


def str_time(value, _format='%H:%M:%S.%f'):
    return value.strftime(_format)


def get_extension_from_comment(comment: str) -> str:
    comment = comment.strip()
    res = 'bin'
    if comment:
        ext = comment.split(" ")[-1]
        if len(ext) < 6 and ext.isalnum():
            return ext
    return res


def check_version(v8_version: str, src_version: str) -> None:
    _v8 = v8_version.split(".")
    _src = src_version.split(".")
    if len(_v8) != 3:
        raise AssertionError(f'Не правильный номер версии v8unpack "{v8_version}"')
    if len(_src) != 3:
        raise AssertionError(f'Не правильная версия исходников "{src_version}"')
    if (int(_v8[0]), int(_v8[1])) != (int(_src[0]), int(_src[1])):
        raise AssertionError(
            f"Версия исходников {_src[0]}.{_src[1]} не соответствует версии v8unpack {_v8[0]}.{_v8[1]}")


def get_descent_file_name(file_name, descent):
    if not descent:
        return file_name
    name: list = file_name.split('.')
    name.insert(-1, str(descent))
    return '.'.join(name)


def get_near_descent_file_name(path, file_name, descent):
    name: list = file_name.split('.')
    startswith = '.'.join(name[0:-1])
    endswith = name[-1]
    size = len(name)
    is_without_descent = False
    try:
        entities = os.listdir(path)
    except FileNotFoundError:
        makedirs(path, exist_ok=True)
        entities = []
    descents = []
    for entity in entities:
        if entity.startswith(startswith) and entity.endswith(endswith):
            _entity = entity.split('.')
            if entity == file_name:
                is_without_descent = True
                continue

            if len(_entity) - 1 != size:
                continue
            full_path = os.path.join(path, entity)
            if os.path.isfile(full_path):
                try:
                    descents.append(int(_entity[-2]))
                except ValueError:  # если во втором разряде не число, значит не наш вариант
                    pass
    if not descents:
        if is_without_descent:
            return path, file_name
        return '', ''
    descents = sorted(descents, reverse=True, key=lambda x: 0 if x > descent else x)
    if descents[0] > descent:
        if is_without_descent:
            return path, file_name
        return '', ''
    return path, f'{startswith}.{descents[0]}.{endswith}'


def remove_descent_from_filename(file_name):
    _name = file_name.split('.')
    try:
        if len(_name) < 3:
            return file_name
        _descent = _name.pop(-2)
        if str(int(_descent)) == _descent:
            return '.'.join(_name)
    except Exception:
        return file_name


def makedirs(name, exist_ok=False):
    for i in range(3):
        try:
            os.makedirs(name, exist_ok=exist_ok)
            return
        except PermissionError:
            time.sleep(0.5)
    raise PermissionError(name)


class FuckingBrackets(ExtException):
    pass


def update_dict(*args):
    size = len(args)
    if size == 1:
        return _update_dict({}, args[0])
    elif size > 1:
        result = args[0]
        for i in range(size - 1):
            result = _update_dict(result, args[i + 1])
        return result


def _update_dict(base, new, _path=''):
    if not new:
        return base
    for element in new:
        try:
            if element in base and base[element] is not None:
                if isinstance(new[element], dict):
                    if isinstance(base[element], dict):
                        base[element] = _update_dict(base[element], new[element], f'{_path}.{element}')
                    else:
                        raise ExtException(
                            message='type mismatch',
                            detail=f'{type(base[element])} in {_path}.{element}',
                            dump={'value': str(base[element])}
                        )
                elif isinstance(new[element], list):
                    base[element].extend(new[element])
                    # base[element] = ArrayHelper.unique_extend(base[element], new[element])
                else:
                    base[element] = new[element]
            else:
                try:
                    base[element] = copy.deepcopy(new[element])
                except TypeError as e:
                    if not base:
                        base = {
                            element: copy.deepcopy(new[element])
                        }
                    else:
                        raise NotImplementedError()
        except ExtException as err:
            raise ExtException(parent=err) from err
        except Exception as err:
            raise ExtException(
                parent=err,
                action='Helper.update_dict',
                detail='{0}({1})'.format(err, _path),
                dump={
                    'element': element,
                    'message': str(err)
                })
    return base


def load_json(filename):
    try:
        with open(filename, 'r', encoding='utf-8') as f:
            data = json.load(f)
        if not isinstance(data, dict):
            raise Exception(f'Index file not dict ({filename})\n')
        return data
    except FileNotFoundError:
        raise Exception(f'Index file not found ({filename})')
    except Exception as err:
        raise Exception(f'Bad index file ({filename}) - {err}\n')


def check_index(index_filename):
    if index_filename:
        index = []
        _index = load_json(index_filename)
        sub_index = _index.pop('index.json', None)
        if sub_index:
            for elem in sub_index:
                index.append(load_json(elem))
        index.append(_index)
        data = update_dict(*index)
        return data
    return None


def get_options_param(options, param_name, default=None):
    try:
        return options[param_name]
    except (KeyError, TypeError):
        return default


def set_options_param(options, param_name, param_value):
    if options is None:
        options = {}
    options[param_name] = param_value
    return options



from .Form803 import Form803


class DocumentForm803(Form803):
    pass
    # def encode_header(self):
    #     return [[
    #         "1",
    #         [
    #             "0",
    #             self.encode_header_title(),
    #         ],
    #         "0"
    #     ]]



import json
import os
import shutil
from base64 import b64decode
from enum import Enum

from . import helper
from .ext_exception import ExtException


class Mode(Enum):
    READ_PARAM = 1
    BEGIN_READ_STRING_VALUE = 2
    BEGIN_READ_MULTI_STRING_VALUE = 6
    END_READ_MULTI_STRING_VALUE = 7
    END_READ_STRING_VALUE = 3
    READ_B64 = 4
    READ_TEXT_FILE = 5


class JsonContainerDecoder:
    def __init__(self, *, src_dir=None, file_name=None):
        self.data = None
        self.raw_data = None
        self.mode = Mode.READ_PARAM
        self.current_object = None
        self.current_value = None
        self.previous_char = None
        self.src_dir = src_dir
        self.file_name = file_name
        self.path = []
        self.params_in_line = 0
        self.line_number = None

    @classmethod
    def decode_mp(cls, params):
        return cls.decode(*params)

    @classmethod
    def decode(cls, params):
        src_dir, file_name, dest_dir = params
        _src_path = os.path.join(src_dir, file_name)
        self = cls(src_dir=src_dir, file_name=file_name)
        encoding = None
        read_as_byte = True
        for code_page in ['utf-8-sig', 'windows-1251']:
            try:
                with open(os.path.join(src_dir, file_name), 'r', encoding=code_page) as entry_file:  # replace BOM
                    try:
                        if entry_file.read(1) == '{':
                            entry_file.seek(0)
                            json.load(entry_file)  # если в файле чистый json воспринимаем его как бинарный файл
                            read_as_byte = True
                        else:
                            read_as_byte = True
                    except json.JSONDecodeError as err:
                        entry_file.seek(0)
                        data = self.decode_file(entry_file)
                        read_as_byte = False
                        encoding = code_page
                        break
            except UnicodeDecodeError:
                continue
            #     try:
            #         with open(os.path.join(src_dir, file_name), 'r', encoding='windows-1251') as entry_file:  # replace BOM
            #             data = self.decode_file(entry_file)
            #             encoding = 'windows-1251'
            #     except UnicodeDecodeError:
            #         read_as_byte = True
            except BigBase64:
                shutil.copy2(os.path.join(src_dir, file_name), os.path.join(dest_dir, file_name + '.c1b64'))
                return
            except Exception as err:
                raise ExtException(parent=err, message=f'Json decode {file_name} error: {err}')
        if read_as_byte:
            shutil.copy2(os.path.join(src_dir, file_name), os.path.join(dest_dir, file_name + '.bin'))
            # with open(os.path.join(src_dir, file_name), 'rb') as entry_file:
            #     data = entry_file.read()
        else:
            if isinstance(data, list):
                with open(f'{os.path.join(dest_dir, file_name)}.json', 'w', encoding='utf-8') as entry_file2:
                    json.dump(data, entry_file2, ensure_ascii=False, indent=2)
            elif isinstance(data, str):
                if encoding is None:
                    encoding = helper.detect_by_bom(_src_path, 'utf-8')
                with open(f'{os.path.join(dest_dir, file_name)}.txt', 'w', encoding=encoding) as entry_file2:
                    entry_file2.write(data)
            # elif isinstance(data, bytes):
            #     with open(f'{os.path.join(dest_dir, file_name)}.bin', 'wb') as entry_file2:
            #         entry_file2.write(data)
            else:
                raise Exception(f'Не поддерживаемый тип данных {type(data)}')

    def decode_file(self, file):
        self.mode = Mode.READ_PARAM
        self.data = []
        self.line_number = 1
        for line in file:
            try:
                self.decode_line(line)
                self.line_number += 1
            except BigBase64 as err:
                raise err from err
            except Exception as err:
                raise ExtException(
                    parent=err,
                    message="Ошибка при разборе скобкофайла",
                    detail=f'{os.path.basename(self.src_dir)}/{self.file_name} проблема до строки {self.line_number}',
                    dump=dict(
                        mode=self.mode,
                        current_object=self.current_object,
                        path=self.path
                    ))
        if self.mode != Mode.READ_PARAM:
            raise ExtException(
                message="Ошибка при разборе скобкофайла",
                detail=f'{os.path.basename(self.src_dir)}/{self.file_name} файл закончен в режиме {self.mode}, '
                       f'создайте проблему на github, укажите текст ошибки и приложите указанный файл'
                # dump=dict(
                #     mode=self.mode,
                #     current_object=self.current_object,
                #     path=self.path
                # )
            )
        if not self.data:
            return ''
        return self.data

    def decode_line(self, line):
        handler = getattr(self, f'_decode_line_{self.mode.name.lower()}')
        return handler(line)

    def _decode_line_read_b64(self, line):
        if line == '\n':
            return
        self.decode_b64_line(line, 0)

    def _decode_line_read_text_file(self, line):
        self.data += line

    def _decode_line_read_param(self, line):
        if line[0] == '{':  # новый объект, исходим из того, что формат записи предполагает только один новый объект
            if self.current_object is None:
                self.current_object = []
                self.data.append(self.current_object)
                self.path.append(self.current_object)
            else:
                self.current_object.append([])
                self.current_object = self.current_object[-1]
                self.path.append(self.current_object)

            self.current_value = ''
            if line.startswith('{#base64'):
                # это скорее всего файл целиком из двоичных данных
                if len(self.data) == 1 and len(self.data[0]) == 2 and self.data[0][0] == '1':
                    raise BigBase64()
                self.mode = Mode.READ_B64
                self.decode_b64_line(line, 1)
            else:
                self.decode_object(line[1:])
        elif line[0] == '}':
            self._end_current_object()
            self.decode_object(line[1:])
        elif line[0] == '\n' and self.current_value and len(self.current_value) == 64:
            # self.current_value = self.current_value
            self.mode = Mode.READ_B64
            return
        else:
            if not self.data and self.current_value is None:
                if line == '\n':
                    return
                self.data = line
                self.mode = Mode.READ_TEXT_FILE
            elif self.data == [[]] and self.current_value == '':  # текстовый файл с json
                self.data = '{\n' + line
                self.mode = Mode.READ_TEXT_FILE
            else:
                raise ExtException(
                    message='Неожиданное начало объекта',
                    detail=f'в файле :{self.src_dir}/{self.file_name}, path:{self.path})')

    def _decode_line_begin_read_string_value(self, line):
        self.decode_object(line)

    def decode_object(self, line):
        for i, char in enumerate(line):
            if self.mode == Mode.READ_PARAM:
                if char == ',':
                    self._end_value()
                elif char == '}':
                    self._end_current_object()
                elif char == '"':
                    if line.endswith(',"\n') and i == len(line) - 2:
                        self.mode = Mode.BEGIN_READ_MULTI_STRING_VALUE
                        self._add_to_current_value(line[i:])
                        break
                    else:
                        self.mode = Mode.BEGIN_READ_STRING_VALUE
                        self._add_to_current_value(char)
                    pass
                elif char == '\n':
                    break
                else:
                    self._add_to_current_value(char)
            elif self.mode == Mode.BEGIN_READ_STRING_VALUE:
                if char == '"':
                    self.mode = Mode.END_READ_STRING_VALUE
                    self._add_to_current_value(char)
                else:
                    self._add_to_current_value(char)
            elif self.mode == Mode.END_READ_STRING_VALUE:
                if char == '"':
                    self.mode = Mode.BEGIN_READ_STRING_VALUE
                    self._add_to_current_value(char)
                elif char == ',':
                    self._end_value()
                elif char == '}':
                    self._end_current_object()
                else:
                    self._add_to_current_value(char)
            else:
                raise NotImplemented(f'mode {self.mode}')
        # if self.current_value:
        #     self.current_object.append(self.current_value)
        #     self.current_value = ''

    def decode_b64_line(self, line, start_pos):
        end_pos = line.find('}')
        if end_pos >= 0:
            self.current_value += line[start_pos:end_pos]
            if start_pos == 1:  # b64 не зарбит на строки
                self.current_value = "#" + self.current_value
            self._end_current_object()
            self.decode_object(line[end_pos + 1:])
            return True
        else:
            self.current_value += line[start_pos:-1]
            return False

    def _decode_line_begin_read_multi_string_value(self, line):
        if line.count('""') * 2 != line.count('"'):
            self.mode = Mode.BEGIN_READ_STRING_VALUE
            self.decode_object(line)
        else:
            self.current_value += line
            return False

    def _decode_line_end_read_multi_string_value(self, line):
        if line[0] in ['{', '}']:
            self._end_value()
            self._decode_line_read_param(line)
        else:
            self.mode = Mode.BEGIN_READ_MULTI_STRING_VALUE
            self.current_value += ",\n"
            self._decode_line_begin_read_multi_string_value(line)

    def _add_to_current_value(self, value):
        try:
            self.current_value += value
        except TypeError:
            self.current_value = value
        self.previous_char = value

    def _end_current_object(self):
        self._end_value()
        if self.path:  # могут быль лишние закрывающие скобки
            self.path.pop()
        self.current_object = self.path[-1] if self.path else None
        self.current_value = None
        self.previous_char = '}'

    def _end_value(self):
        self.mode = Mode.READ_PARAM
        if self.previous_char != '}':
            self.current_object.append(self.current_value)
            self.current_value = ''
        self.previous_char = ','

    @classmethod
    def encode_mp(cls, params):
        return cls.encode(*params)

    @classmethod
    def encode(cls, params):
        src_dir, file_name, dest_dir = params
        try:
            # try:
            with open(os.path.join(src_dir, file_name), 'r', encoding='utf-8') as file:
                data = json.load(file)
            self = cls(src_dir=src_dir, file_name=file_name)
            raw_data = self.encode_root_object(data)
            if dest_dir is not None:
                cls.write_data(dest_dir, file_name[:-5], raw_data)
            return raw_data
        except Exception as err:
            raise ExtException(parent=err, detail=file_name) from err

    def encode_root_object(self, data):
        raw_data = ''
        for elem in data:
            if raw_data:
                raw_data += ',\n'
            raw_data += self.encode_object(elem, True)
        return raw_data

    def encode_object(self, data, first=True):
        if first:
            raw_data = '{'
        else:
            raw_data = '\n{'
            self.params_in_line = 0
        data_len = len(data)
        for i in range(data_len):
            elem = data[i]
            if isinstance(elem, list):
                raw_data += self.encode_object(elem, False)
                if i == data_len - 1:
                    raw_data += '\n'
                    self.params_in_line = 0
                pass
            elif isinstance(elem, str):
                if elem.startswith('#base64:'):
                    j = 72
                    raw_data += f'{elem[0:j]}'
                    while j <= len(elem):
                        raw_data += f'\r\n{elem[j: j + 64]}'
                        j += 64
                elif elem.startswith('##base64:'):  # b64 в одну строку
                    raw_data += elem[1:]
                else:
                    j = len(elem)
                    is_base64 = False
                    if j > 64:  # base64 нужно переносить
                        try:
                            b64decode(elem)
                            is_base64 = True
                        except ValueError:
                            pass

                    if is_base64:
                        _first = True
                        while j:
                            if _first:
                                _first = False
                            else:
                                raw_data += '\r\n'

                            if j > 64:
                                raw_data += elem[:64]
                                elem = elem[64:]
                                j = len(elem)
                            else:
                                raw_data += elem
                                break
                    else:
                        raw_data += elem
            elif elem is None:
                pass
            else:
                raise NotImplementedError(elem)

            if i != data_len - 1:
                raw_data += ','
                self.params_in_line += 1

        raw_data += '}'
        return raw_data
        pass

    @staticmethod
    def write_data(dest_dir, file_name, data):
        with open(os.path.join(dest_dir, file_name), 'w', encoding='utf-8-sig') as f:  # replace BOM
            return f.write(data)


def json_decode(src_dir, dest_dir, *, pool=None):
    """
    Рекурсивно добавляет файлы из директории в контейнер

    :param src_dir: каталог файлов, результатов работы v8unpack -U
    :type src_dir: string
    :param dest_dir: каталог файлов, в который будут помещены json файлы
    :type dest_dir: string
    :param pool: пул процессов для параллельного выполнения, если не передан будет поднят свой
    :type: pool: Multiprocessing Pool
    """
    tasks = []
    _decode(src_dir, dest_dir, tasks)
    helper.run_in_pool(JsonContainerDecoder.decode, tasks, pool=pool, title='Конвертируем скобкофайл в json',
                       need_result=False)


def _decode(src_dir, dest_dir, tasks):
    entries = sorted(os.listdir(src_dir))
    helper.clear_dir(dest_dir)
    for entry in entries:
        src_entry_path = os.path.join(src_dir, entry)
        dest_entry_path = os.path.join(dest_dir, entry)
        if os.path.isdir(src_entry_path):
            try:
                helper.makedirs(dest_entry_path, exist_ok=True)
            except Exception as err:
                raise ExtException(
                    parent=err, message='Ошибка при создании папки',
                    detail=f'{dest_entry_path} ({err})', action='json_decode')
            try:
                _decode(src_entry_path, dest_entry_path, tasks)
            except Exception as err:
                raise ExtException(parent=err, detail=f'{entry} {src_dir}', action='json_decode')
        else:
            tasks.append((src_dir, entry, dest_dir))


def json_encode(src_dir, dest_dir, *, pool=None):
    """
    Распаковка контейнера. Сахар для ContainerReader

    :param src_dir: каталог файлов, результатов работы v8unpack -U
    :type src_dir: string
    :param dest_dir: каталог файлов, в который будут помещены json файлы
    :type dest_dir: string
    :param pool: пул процессов для параллельного выполнения, если не передан будет поднят свой
    :type: pool: Multiprocessing Pool
    """
    tasks = []
    helper.clear_dir(dest_dir)
    _encode(src_dir, dest_dir, tasks)
    helper.run_in_pool(JsonContainerDecoder.encode, tasks, pool=pool, title='Конвертироуем json в скобковайл',
                       need_result=False)


def _encode(src_dir, dest_dir, tasks):
    entries = os.listdir(src_dir)
    for entry in entries:
        try:
            src_entry_path = os.path.join(src_dir, entry)
            dest_entry_path = os.path.join(dest_dir, entry)
            if os.path.isdir(src_entry_path):
                os.mkdir(dest_entry_path)
                _encode(src_entry_path, dest_entry_path, tasks)
            else:
                extension = entry.split('.')[-1]
                if extension == 'txt':
                    encoding = helper.detect_by_bom(src_entry_path, 'utf-8')
                    try:
                        with open(src_entry_path, 'r', encoding=encoding) as entry_file:
                            with open(dest_entry_path[:-3], 'w', encoding=encoding) as f:
                                f.write(entry_file.read())
                    except UnicodeDecodeError:
                        encoding = 'windows-1251'
                        with open(src_entry_path, 'r', encoding=encoding) as entry_file:
                            with open(dest_entry_path[:-3], 'w', encoding=encoding) as f:
                                f.write(entry_file.read())
                elif extension == 'bin' or extension == 'c1b64':
                    shutil.copy2(
                        os.path.join(src_dir, entry),
                        os.path.join(dest_dir, os.path.splitext(entry)[0])
                    )
                    # with open(src_entry_path, 'rb') as entry_file:
                    #     with open(dest_entry_path[:-3], 'wb') as f:
                    #         f.write(entry_file.read())
                else:
                    tasks.append((src_dir, entry, dest_dir))
        except Exception as err:
            raise ExtException(parent=err, detail=f'{entry} {src_dir}', action='json_encode')


class BigBase64(Exception):
    pass



from ..MetaDataObject.core.Container import Container


class ExternalDataSourceCube(Container):
    ext_code = {'obj': 2, 'mgr': 1}

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][1][1]



# -*- coding: utf-8 -*-
import os
import zlib
from datetime import datetime

from . import helper
from .container import Container, Container64
from .ext_exception import ExtException


def extract(filename, folder, deflate=True, recursive=True):
    """
    Распаковка контейнера. Сахар для ContainerReader

    :param filename: полное имя файла-контейнера
    :type filename: string
    :param folder: каталог назначения
    :type folder: string
    :param deflate: паспаковка
    :type deflate: boolean
    :param recursive: рекурсивно достаем все контейнеры
    :type recursive: boolean
    """
    begin = datetime.now()
    print(f'{"Распаковываем бинарник":30}:', end="")
    helper.clear_dir(folder)
    with open(filename, 'rb') as f:
        offset = 0
        container_index = 0
        while True:
            try:
                container = detect_format(f, offset)
                container.read(f, offset)
                container.extract(os.path.join(folder, str(container_index)), deflate, recursive,
                                  progress=container_index)
                container_index += 1
                offset += container.size
                if offset == 0:
                    raise NotImplementedError()
            except EOFError:
                return
            except Exception as err:
                raise err from err

    print(f"{datetime.now() - begin}")


def detect_format(f, offset):
    f.seek(offset)
    first = f.read(8)
    if first[0:4] == b'\xFF\xFF\xFF\x7F':
        return Container()
    elif first == b'':
        raise EOFError()
    # elif first == b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF':
    else:
        return Container64()
        # raise Exception('Файл не является контейнером')


def decompress_and_extract(src_folder, dest_folder, *, pool=None):
    helper.clear_dir(dest_folder)
    containers = os.listdir(src_folder)
    helper.clear_dir(dest_folder)
    tasks = []
    for container in containers:
        _src_folder = os.path.join(src_folder, container)
        _dest_folder = os.path.join(dest_folder, container)
        helper.clear_dir(_dest_folder)
        entries = os.listdir(_src_folder)
        for filename in entries:
            tasks.append([_src_folder, filename, _dest_folder])

    helper.run_in_pool(decompress_file_and_extract, tasks, pool=pool, title=f'{"Распаковываем контейнеры":30}')


def decompress_file_and_extract(params):
    src_folder, filename, dest_folder = params
    src_path = os.path.join(src_folder, filename)
    dest_path = os.path.join(dest_folder, filename)
    file_is_container = None

    # wbits = -15 т.к. у архивированных файлов нет заголовков
    decompressor = zlib.decompressobj(-15)
    try:
        with open(dest_path, 'wb') as dest:
            with open(src_path, 'rb') as src:
                while True:
                    buf = decompressor.unconsumed_tail
                    if buf == b'':
                        buf = src.read(8192)
                        if buf == b'':
                            break
                    data = decompressor.decompress(buf)
                    if file_is_container is None:
                        file_is_container = data[0:4] == b'\xFF\xFF\xFF\x7F'
                    if data == b'':
                        break
                    dest.write(data)

        # Каждый файл внутри контейнера может быть контейнером
        # Для проверки является ли файл контейнером проверим первые 4 бита
        # Способ проверки ненадежный - нужно придумать что-то другое

        if file_is_container:
            temp_filename = dest_path + ".temp"
            os.rename(dest_path, temp_filename)
            with open(temp_filename, 'rb') as f:
                container = Container()
                container.read(f)
                container.extract(dest_path, recursive=True)
            os.remove(temp_filename)

    except Exception as err:
        raise ExtException(
            parent=err, message="Ошибка при разархифировании контейнера",
            detail=f'{filename} ({err})')



import os
import re

from ..core.Simple import SimpleNameFolder
from ... import helper
from ...ext_exception import ExtException
from .Form802Elements.FormElement import FormElement as FormElement802, FormProps as FormProps802
from .Form803Elements.FormElement import FormElement as FormElement803, FormProps as FormProps803, calc_offset
from ... import helper
from ...ext_exception import ExtException


class Form8x(SimpleNameFolder):
    _obj_name = "Form"
    double_quotes = re.compile(r'("")')
    quotes = re.compile(r'(")')

    def __init__(self, *, obj_name=None, options=None):
        super().__init__(obj_name=obj_name, options=options)
        self.form = []
        self.elements = []
        self.elements_data = {}
        self.props = []

    def decode_object(self, src_dir, uuid, dest_dir, dest_path, version, header_data):
        self.set_header_data(header_data)
        self.set_write_decode_mode(dest_dir, dest_path)
        self.decode_data(src_dir, uuid)
        self.decode_code(src_dir)

    def decode_form0(self, src_dir, uuid):
        try:
            form = helper.brace_file_read(src_dir, f'{uuid}.0')
        except FileNotFoundError:
            self.form.append([])
            return
        try:
            _code = helper.str_decode(self.getset_form_code(form, 'Код в отдельном файле', self.header))
            if _code is not None:
                _code = self.double_quotes.sub(r'"', _code)
                self.code['obj'] = _code
                self.header['code_info_obj'] = 'Код в отдельном файле'
        except Exception as err:
            raise ExtException(parent=err, detail=self.header['uuid'])
        self.form.append(form)

    @classmethod
    def getset_form_code(cls, form, new_value=None, header=None):
        err_detail = f'{header["uuid"]} {header["name"]} ' \
                     f'опытным путем подобрано, если у Вас код не достается' \
                     f'обновитесь до последней версии, и если не поможет создайте issue с дампом'
        len_form_0 = len(form[0])
        if len_form_0 > 2 and form[0][0] in ['4', '3', '2']:
            code = form[0][2]
            if not isinstance(code, str):
                raise ExtException(
                    message='Not supported forms',
                    detail=err_detail,
                    dump=form
                )
            form[0][2] = new_value
            if len_form_0 != 10:
                a = 1
            return code

        last_level = cls.get_last_level_array(form)
        if len_form_0 < 10 and (last_level[0] == '49' or last_level[0] == '4'):
            return ''

        if len(last_level) > 10 \
                and last_level[0] in ['22', '1'] \
                and last_level[-1] == '0' \
                and last_level[-2] == '0':
            code_index = -8
            code = last_level[code_index]
            if not isinstance(code, str):
                raise ExtException(
                    message='Not supported forms',
                    detail=err_detail,
                    dump=form
                )
            if new_value is not None:
                last_level[code_index] = new_value
            return code
        return ''

    def decode_old_form(self, src_dir):
        file_name = f'{self.header["uuid"]}.0'
        _code_dir = os.path.join(src_dir, file_name)
        if os.path.exists(_code_dir):
            if os.path.isdir(_code_dir):
                self.form.append(helper.brace_file_read(_code_dir, 'form'))
                try:
                    encoding = helper.detect_by_bom(os.path.join(_code_dir, 'module'), 'utf-8')
                    self.code['obj'] = self.read_raw_code(_code_dir, 'module', encoding=encoding)
                except UnicodeDecodeError:
                    encoding = 'windows-1251'
                    self.code['obj'] = self.read_raw_code(_code_dir, 'module', encoding=encoding)
                self.header['code_encoding_obj'] = encoding  # можно безболезненно поменять на utf-8-sig
                self.header[f'code_info_obj'] = 1
            else:
                self.decode_form0(src_dir, self.header["uuid"])
                pass

    def decode_old_elements(self):
        try:
            if not self.form[0]:
                return
            _ver = self.form[0][0][0]
            if _ver == '2':
                self.elements = FormElement803.decode_list(self, self.form[0][0][1], 23)
                # self.props = FormPros802.decode_list(self, self.form[0][0][2][2])
            else:
                self.elements = FormElement802.decode_list(self, self.form[0][0][1][2][2])
                self.props = FormProps802.decode_list(self, self.form[0][0][2][2])
        except Exception as err:
            pass  # todo если какие то елементы формы не разбираются, не прерываем
            # raise ExtException(parent=err)

    def write_decode_object(self, dest_dir, dest_path, file_name):
        super(Form8x, self).write_decode_object(dest_dir, dest_path, file_name)
        helper.json_write(self.form, self.new_dest_dir, f'{file_name}.form{self.version}.json')
        if self.elements:
            helper.json_write(self.elements, self.new_dest_dir, f'{file_name}.elements.tree{self.version}.json')
            helper.json_write(self.elements_data, self.new_dest_dir, f'{file_name}.elements.data{self.version}.json')
        if self.props:
            helper.json_write(self.props, self.new_dest_dir, f'{file_name}.props{self.version}.json')
        return []

    def decode_data(self, src_dir, uuid):
        raise NotImplemented()

    def get_decode_header(self, header):
        return self.get_decode_obj_header(header)[1][1]

    def get_decode_obj_header(self, header_data):
        return header_data[0][1] if self.obj_version == '0' else header_data[0][1][1]

    def encode_object(self, src_dir, file_name, dest_dir):
        super(Form8x, self).encode_object(src_dir, file_name, dest_dir)
        try:
            self.form = helper.json_read(src_dir, f'{file_name}.form{self.version}.json')
        except FileNotFoundError:
            self.form = self.encode_empty_form()
        self.encode_data()
        self.encode_nested_includes(src_dir, file_name, dest_dir, None)

    def encode_old_elements(self, src_dir, file_name, dest_dir, parent_id):
        def get_form2_elem_index(_root_data, _file_name):
            try:
                index_command_panel_count = calc_offset([(21, 0)], _root_data)
                command_panel_count = int(_root_data[index_command_panel_count])
                index_root_elem_count = index_command_panel_count + command_panel_count + 1
                return index_root_elem_count, index_command_panel_count
            except Exception as err:
                raise ExtException(
                    message='случай требующий анализа, предоставьте образец формы разработчикам',
                    detail=f'{self.header["name"]} {_file_name}, {err}')
        try:
            if not self.form[0]:
                return
            _ver = self.form[0][0][0]
            if _ver == '2':
                root_data = self.form[0][0][1]
                index = get_form2_elem_index(root_data, file_name)
                index_root_element_count = index[0]
                if root_data[index_root_element_count] == 'Дочерние элементы отдельно':
                    self.elements = helper.json_read(src_dir, f'{file_name}.elements.tree{self.version}.json')
                    self.elements_data = helper.json_read(src_dir, f'{file_name}.elements.data{self.version}.json')
                    # root_data[index_root_element_count] = str(len(self.elements))
                    FormElement803.encode_list(self, self.elements, root_data, index_root_element_count)
            else:
                FormElement802.encode_list(self, src_dir, file_name, self.version, self.form[0][0][1][2][2])
                FormProps802.encode_list(self, src_dir, file_name, self.version, self.form[0][0][2][2])
        except Exception as err:
            raise ExtException(parent=err, message='Ошибка при разборе формы', detail=f'{file_name}')

    def encode_header(self):
        return self.header['data']

    def encode_nested_includes(self, src_dir, file_name, dest_dir, parent_id):
        raise NotImplemented()

    def encode_data(self):
        raise NotImplemented()

    def encode_empty_form(self):
        raise NotImplemented()

    def encode_write(self, dest_dir):
        raise NotImplemented()

    def write_encode_code(self, dest_dir):
        pass

    def write_old_encode_object(self, dest_dir):
        helper.brace_file_write(self.encode_header(), dest_dir, self.header["uuid"])
        self.file_list.append(self.header["uuid"])
        if self.header.get('code_info_obj'):
            dir_name = f'{self.header["uuid"]}.0'
            _code_dir = os.path.join(dest_dir, dir_name)
            self.file_list.append(dir_name)
            helper.makedirs(_code_dir, exist_ok=True)
            helper.brace_file_write(self.form[0], _code_dir, 'form')
            _encoding = self.header.get('code_encoding_obj', 'utf-8-sig')
            self.write_raw_code(self.code.get('obj', ''), _code_dir, 'module', encoding=_encoding)

    @classmethod
    def get_last_level_array(cls, data):
        while True:
            if isinstance(data[-1], list):
                return cls.get_last_level_array(data[-1])
            else:
                return data



from ..MetaDataObject.core.Container import Container


class CalculationRegisterRecalculations(Container):
    pass

    @classmethod
    def get_decode_header(cls, header):
        return header[0][1][7][1]



from ..MetaDataObject.core.IncludeSimple import IncludeSimple


class ReportCommand(IncludeSimple):
    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][2][9]



from ..MetaDataObject.core.Simple import SimpleNameFolder
from ..MetaDataObject.versions.CommonForm803 import CommonForm803


class CommonForm(SimpleNameFolder):
    versions = {
        '803': CommonForm803,
        4: CommonForm803
    }



import json
import sys
import traceback
from os import path


# 3xxx - Ошибки в нашем коде
# 4xxx - Ошибки вызванные некорректными действиями пользователей
# 5xxx - Ошибки вызванные временной недоступностью функционала,повторный вызов с теми же параметрами может быть успешным
# 6xxx - Ошибки во внешних сервисах и библиотеках
# 8xxx - Запросы не реализованного, но потонциально интересного функционала
# 9xxx - Обертка над стандартными ошибками для испльзования в коде


class ExtException(Exception):
    _code = 3000
    _http_code = 500
    _message = None

    # _message = "Unknown error"

    def __new__(cls, *args, **kwargs):
        parent = kwargs.get('parent')
        if parent:
            # прокидываем ошибку наверх
            if isinstance(parent, ExtException) \
                    and parent.__class__ != ExtException \
                    and kwargs.get('message') is None \
                    and cls == ExtException:
                return super().__new__(parent.__class__)
            if isinstance(parent, dict):  # пытаемся поднять правильный класс
                class_name = parent.pop('__name__', None)
                module_name = parent.pop('__module__', None)
                if class_name and module_name:
                    try:
                        m = __import__(module_name)
                        parts = f'{module_name}.{class_name}'.split('.')
                        for comp in parts[1:]:
                            m = getattr(m, comp)
                        return super().__new__(m)
                    except:
                        pass
        return super().__new__(cls)

    def __init__(self, *, parent=None, code=None, message=None, detail=None, action=None, dump=None,
                 skip_traceback=0, stack=None):
        self.skip_traceback = skip_traceback
        self.code = code
        self.message = message
        self.detail = detail

        self.dump = {} if dump is None else dump
        self.stack = [] if stack is None else stack

        if action and not isinstance(action, str):  # это класс действте
            # self.add_action_to_stack(action)
            self.action = action.name
        else:
            self.action = action

        self.new_msg = bool(message)
        if parent:
            if isinstance(parent, ExtException):  # прокидываем ошибку наверх
                self.add_parent_to_stack(parent)
                if not message:
                    self.message = parent.message
                    self.detail = parent.detail
                    self.code = parent.code
                    self.dump = parent.dump
            elif isinstance(parent, Exception):
                if not message:
                    self.message = 'Unknown Error'
                if not detail:
                    self.detail = str(parent)
                self.add_sys_exc_to_stack()
            elif isinstance(parent, dict):
                self.code = parent.get('code', self._code)
                self.message = parent.get('message')
                self.detail = parent.get('detail')
                self.dump = parent.get('dump', {})
                self.stack = parent.get('stack', [])
                self.action = parent.get('action')

            else:
                raise NotImplementedError(f'ExtException parent={type(parent)}')
        if not self.stack:
            self.add_sys_exc_to_stack()
        if not self.message:
            self.message = self._message
        pass

    def add_action_to_stack(self, action):
        try:
            self.action = action.name
            if action.stat:
                action.set_end('Error')
                self.dump['_action_stat'] = action.stat
        except:
            pass

    def add_parent_to_stack(self, parent):
        if not isinstance(parent, ExtException):
            return None
        self.stack += parent.stack
        if not parent.action:
            return
        if not self.stack:
            data = self.get_sys_exc_info()
            if data:
                parent.dump['traceback'] = data['traceback']

        _stack = {
            'action': parent.action
        }
        if parent.new_msg:
            _stack['message'] = parent.message
            if parent.detail:
                _stack['detail'] = parent.detail
        if parent.dump:
            _stack['dump'] = parent.dump
        self.stack.append(_stack)

    def add_sys_exc_to_stack(self):
        try:
            data = self.get_sys_exc_info()
        except IndexError:
            return
        if not data:
            return
        data['action'] = self.action
        self.stack.append(data)

    def get_sys_exc_info(self):
        exc_info = sys.exc_info()
        if exc_info[2] is None:
            return None
        last_call = traceback.extract_tb(exc_info[2], limit=self.skip_traceback + 1)
        last_call = last_call[self.skip_traceback]
        if isinstance(exc_info[1], ExtException):
            message = getattr(exc_info[1], 'message')
            detail = getattr(exc_info[1], 'detail')
        else:
            message = exc_info[0].__name__
            detail = str(exc_info[1])
        return {
            'message': message,
            'detail': detail,
            'traceback': f'{path.basename(path.dirname(last_call.filename))}/{path.basename(last_call.filename)}, '
                         f'{last_call.name}, line {last_call.lineno}'
        }

    @property
    def title(self):
        title = ''
        if self.code:
            title += f'{self.code}:'

        title += f'{self.__class__.__name__} {self.message}'

        if self.detail:
            title += f' - {self.detail}'
        return title

    def __str__(self):
        res = f'{self.title}'
        if self.dump:
            res += f'\nDump: action={self.action}; '
            for elem in self.dump:
                res += f'{elem}={self.dump[elem]}; '
        if self.stack:
            res += '\n Stack:'
            for stack in self.stack:
                _action = stack.get('action', '')
                res += f'\n  - {_action}: '
                for elem in stack:
                    if elem != 'action':
                        res += f'{elem}={stack[elem]}; '
        return res

    def dumps(self):
        return json.dumps(self.to_dict(), ensure_ascii=False)

    def to_dict(self):
        return {
            '__name__': self.__class__.__name__,
            '__module__': self.__class__.__module__,
            'message': self.message,
            'detail': self.detail,
            'code': self.code,
            'action': self.action,
            'dump': self.dump,
            'stack': self.stack
        }

    @property
    def http_code(self):
        return self._http_code


class HandlerNotFoundError(ExtException):
    _code = 3100
    _message = "Handler not found",


class UserError(ExtException):
    _code = 4000


class Unauthorized(ExtException):
    _code = 4011
    _http_code = 401
    _message = "Unauthorized"


class ResourceNotAvailable(ExtException):
    _code = 4100
    _message = "Resource not available"


class KeyNotFound(ExtException):
    _code = 4200
    _message = "Required parameter are missing"


class TooManyRequests(ExtException):
    _code = 4290
    _message = "Too Many Requests"
    # dump
    # X-RateLimit-Limit - количество запросов, которые равномерно можно сделать в течение интервала до появления 429 ошибки
    # X-Lognex-Retry-TimeInterval - интервал в миллисекундах, в течение которого можно сделать эти запросы
    # X-RateLimit-Remaining - Число запросов, которые можно отправить до получения 429 ошибки
    # X-Lognex-Reset - время до сброса ограничения в миллисекундах. Равно нулю, если ограничение не установлено
    # X-Lognex-Retry-After - время до сброса ограничения в миллисекундах.


class NotFound(ExtException):
    _code = 4300
    _message = "Not found"


class WaitingUserAction(ExtException):
    _code = 4911
    _message = "Waiting for user action"


class AccessDenied(ExtException):
    _code = 4031
    _http_code = 403
    _message = "Access denied"


class CancelOperation(ExtException):
    _code = 4999
    _message = "Cancel operation",


class NotAvailable(ExtException):
    _code = 5000
    _message = "Not available",


class ForeignError(ExtException):
    _code = 6000
    _message = "Unknown external error",


class ExtTimeoutError(ExtException):
    _code = 5008
    _http_code = 408
    _message = "Timeout",


class ExtNotImplemented(ExtException):
    _code = 8000
    _http_code = 400
    _message = "Not implemented",


def dumps_error(err):
    return json.dumps({
        '__name__': err.__class__.__name__,
        '__module__': err.__class__.__module__,
        'message': str(err)
    })


import os

from .. import helper
from ..MetaObject.Configuration803 import Configuration803
from ..version import __version__


class ConfigurationExtension803(Configuration803):
    info = ['6', '8', '9']
    ext_code = {
        'con': '5',
        'app': '6',
        'ssn': '7'
    }
    _obj_info = {
        'a': 'a',
    }

    def decode(self, src_dir, dest_dir, *, version=None, **kwargs):
        self.header = {}
        root = helper.brace_file_read(src_dir, 'configinfo')
        self.header['v8unpack'] = __version__
        self.header['file_uuid'] = root[1][1]
        self.header['version'] = root[0][1]
        # self.header['versions'] = root[2]
        self.header['copyinfo'] = root[1]
        # self.header['copyinfo'][2] = b64decode(self.header['copyinfo'][2])[:-16].hex()
        self.header['data'] = helper.brace_file_read(src_dir, f'{self.header["file_uuid"]}')
        _form_header = self.get_decode_header(self.header['data'])
        helper.decode_header(self.header, _form_header)
        product_version = self.header['data'][0][3][1][1][15]
        if version is None:
            self.header['compatibility_version'] = self.header['data'][0][3][1][1][43]
        else:
            self.header['compatibility_version'] = version
            self.header['data'][0][3][1][1][43] = version

        self.decode_code(src_dir)

        for i in self.info:  # хз что это
            file_name = f'{self.header["uuid"]}.{i}'
            if os.path.isdir(os.path.join(src_dir, file_name)):
                continue
            try:
                self.header[f'info{i}'] = helper.brace_file_read(src_dir, file_name)
            except FileNotFoundError:
                pass
        file_name = self.get_class_name_without_version()
        self._decode_info(src_dir, dest_dir, file_name)
        tasks = self.decode_includes(src_dir, dest_dir, '', self.header['data'])

        helper.txt_write(helper.str_decode(product_version), dest_dir, 'version.bin', encoding='utf-8')
        helper.json_write(self.header, dest_dir, f'{self.get_class_name_without_version()}.json')
        self.write_decode_code(dest_dir, self.__class__.__name__)

        return tasks

    def encode(self, src_dir, dest_dir, *, file_name=None, include_index=None, file_list=None):
        self.header = helper.json_read(src_dir, f'{self.get_class_name_without_version()}.json')

        self.set_product_info(src_dir, file_name)

        version = self.get_options('version')
        if version is not None:
            self.header['data'][0][3][1][1][43] = version

        gui = self.get_options('gui')
        if gui is not None:
            self.header['data'][0][3][1][1][38] = gui

        helper.check_version(__version__, self.header.get('v8unpack', ''))

        if include_index:
            self.fill_header_includes(include_index)

        # self.header['copyinfo'][2] = b64encode(bytes.fromhex(self.header['copyinfo'][2]+md5().digest().hex())).decode()
        root = [
            ["0", self.encode_version()],
            self.header['copyinfo'],
            # self.header['versions']
            ["____versions____"]
        ]
        self.encode_code(src_dir, self.__class__.__name__)
        self.write_encode_code(dest_dir)
        _file_name = self.get_class_name_without_version()
        self._encode_info(src_dir, _file_name, dest_dir)
        helper.brace_file_write(root, dest_dir, 'configinfo')
        helper.brace_file_write(self.header['data'], dest_dir, self.header["file_uuid"])
        for i in self.info:  # хз что это
            try:
                helper.brace_file_write(self.header[f'info{i}'], dest_dir, f'{self.header["uuid"]}.{i}')
            except KeyError:
                pass

        return None

    def set_product_version(self, product_version):
        _version = product_version
        if self.product:
            _version = f'{self.product}_{_version}'
        self.header['data'][0][3][1][1][15] = helper.str_encode(_version)



from ...MetaDataObject import MetaDataObject
from ...ext_exception import ExtException


class IncludeSimple(MetaDataObject):
    ext_code = {'obj': 2}

    def __init__(self, *, obj_name=None, options=None):
        super().__init__(obj_name=obj_name, options=options)
        self.new_dest_path = None
        self.new_dest_dir = None

    @classmethod
    def decode(cls, src_dir, file_name, dest_dir, dest_path, version, parent_type=None):
        raise Exception('Так быть не должно, этот класс обслуживает вложенные объекты')

    @classmethod
    def decode_local_include(cls, parent, header_data, src_dir, dest_dir, dest_path, options):
        try:
            self = cls(options=options)
            self.set_header_data(header_data)
            self.set_write_decode_mode(dest_dir, dest_path)
            self.decode_code(src_dir)
            self.write_decode_object(dest_dir, self.new_dest_path, self.new_dest_file_name)
        except Exception as err:
            raise ExtException(
                parent=err,
                action=f'{cls.__name__}.decode_local_include'
            ) from err

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][3][2][9]

    def encode_object(self, src_dir, file_name, dest_dir):
        self.encode_code(src_dir, self.header["name"])
        return []

    def get_include_obj_uuid(self):
        return self.header['data']

    def write_encode_object(self, dest_dir):
        self.write_encode_code(dest_dir)




from ... import helper
from ...MetaDataObject import MetaDataObject
from ...ext_exception import ExtException


class Container(MetaDataObject):
    predefined_file_number = None

    @classmethod
    def get_decode_header(cls, header):
        return header[0][1][1]

    @classmethod
    def get_decode_includes(cls, header_data):
        try:
            return [header_data[0]]
        except IndexError:
            raise ExtException(message='Include types not found', detail=cls.__name__)

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        super().decode_object(src_dir, file_name, dest_dir, dest_path, version, header_data)
        if self.help_file_number is not None:
            self._decode_html_data(src_dir, self.new_dest_dir, self.new_dest_file_name, header_field='help',
                                   file_number=self.help_file_number)
        if self.predefined_file_number is not None:
            self._decode_predefined(src_dir, self.new_dest_dir)
        self.decode_code(src_dir)

    def _decode_predefined(self, src_dir, dest_dir):
        try:
            data = helper.bin_read(src_dir, f'{self.header["uuid"]}.{self.predefined_file_number}')
            helper.bin_write(data, dest_dir, 'Предустановленные данные.bin')
        except FileNotFoundError:
            return

    def set_write_decode_mode(self, dest_dir, dest_path):
        self.set_mode_decode_in_name_folder(dest_dir, dest_path)

    def encode_object(self, src_dir, file_name, dest_dir):
        if self.help_file_number is not None:
            self._encode_html_data(src_dir, file_name, dest_dir, header_field='help', file_number=self.help_file_number)
        if self.predefined_file_number is not None:
            self._encode_predefined(src_dir, dest_dir)
        self.encode_code(src_dir, self.__class__.__name__)
        return []

    def _encode_predefined(self, src_dir, dest_dir):
        try:
            package = helper.bin_read(src_dir, 'Предустановленные данные.bin')
            file_name = f'{self.header["uuid"]}.{self.predefined_file_number}'
            helper.bin_write(package, dest_dir, file_name)
            self.file_list.append(file_name)
        except FileNotFoundError:
            return

    @classmethod
    def encode_get_include_obj(cls, src_dir, dest_dir, include, tasks, options, parent_id, include_index):
        cls.encode_get_include_obj_from_named_folder(src_dir, dest_dir, include, tasks, options, parent_id, include_index)

    def get_encode_file_name(self, file_name):
        return self.get_obj_name()



from .Form801 import Form801
from .Form802 import Form802
from .Form803 import Form803


class OldForm803(Form803):
    def get_decode_obj_header(self, header):
        return header[0]

    # def encode_header(self):
    #     return [[
    #         "1",
    #         self.encode_header_title(),
    #         "0"
    #     ]]


class OldForm802(Form802):
    def get_decode_obj_header(self, header):
        return header[0]

    # def encode_header(self):
    #     return [[
    #         "1",
    #         self.encode_header_title(),
    #         "0"
    #     ]]


class OldForm801(Form801):
    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][1]

    # def get_decode_obj_header(self, header):
    #     return header[0]
    #
    # def encode_header(self):
    #     return [[
    #         "1",
    #         self.encode_header_title(),
    #         "0"
    #     ]]



import os
from base64 import b64encode

from .. import helper
from ..MetaObject import MetaObject
from ..ext_exception import ExtException
from ..metadata_types import MetaDataTypes, MetaDataGroup
from ..version import __version__


class Configuration803(MetaObject):
    ext_code = {
        'seance': 7,
        '803': 6,
        '802': 0,
        'con': 5
    }
    help_file_number = 3

    _images = {
        'Заставка': 2
    }
    _obj_info = {
        '4': '4',
        '9': '9',
        '8': '8',
        'a': 'a',
        'b': 'b',
        'c': 'c',
        'd': 'd',
    }

    def __init__(self, *, obj_name=None, options=None):
        super().__init__(obj_name=obj_name, options=options)
        self.counter = {}
        self.product = None

    def decode(self, src_dir, dest_dir, *, version=None, **kwargs):
        self.header = {}
        root = helper.brace_file_read(src_dir, 'root')
        self.header['v8unpack'] = __version__
        self.header["file_uuid"] = root[0][1]
        self.header["root_data"] = root

        _header_data = helper.brace_file_read(src_dir, f'{self.header["file_uuid"]}')
        self.set_header_data(_header_data)

        file_name = self.get_class_name_without_version()

        self.header['version'] = helper.brace_file_read(src_dir, 'version')
        self.header['versions'] = helper.brace_file_read(src_dir, 'versions')
        # shutil.copy2(os.path.join(src_dir, 'versions.json'), os.path.join(dest_dir, 'versions.json'))

        self.decode_code(src_dir)
        self._decode_html_data(src_dir, dest_dir, 'help', header_field='help', file_number=self.help_file_number)
        self._decode_images(src_dir, dest_dir)
        self._decode_info(src_dir, dest_dir, file_name)

        tasks = self.decode_includes(src_dir, dest_dir, '', self.header['data'])

        helper.json_write(self.header, dest_dir, f'{file_name}.json')
        self.write_decode_code(dest_dir, file_name)
        return tasks

    @classmethod
    def get_decode_header(cls, header):
        return header[0][3][1][1][1][1]

    def decode_includes(self, src_dir, dest_dir, dest_path, header_data):
        tasks = []
        index_includes_group = 2
        count_includes_group = int(header_data[0][index_includes_group])
        for index_group in range(count_includes_group):
            group = header_data[0][index_includes_group + index_group + 1]
            group_uuid = group[0]
            group_version = group[1][0]
            try:
                metadata_group = MetaDataGroup(group_uuid)
            except ValueError:
                raise ExtException(message='Неизвестная группа метаданных', detail=group_uuid)
            include = group[1][1] if group_version == '6' else group[1]
            self.decode_include(src_dir, dest_dir, dest_path, tasks, include)
        return tasks

    def encode(self, src_dir, dest_dir, *, file_name=None, include_index=None, file_list=None):
        _file_name = self.get_class_name_without_version()
        self.header = helper.json_read(src_dir, f'{_file_name}.json')
        helper.check_version(__version__, self.header.get('v8unpack', ''))

        if include_index and self.get_options('auto_include'):
            self.fill_header_includes(include_index)  # todo dynamic index

        helper.brace_file_write(self.header["root_data"], dest_dir, 'root')
        file_list.append('root')
        helper.brace_file_write(self.encode_version(), dest_dir, 'version')
        file_list.append('version')
        # shutil.copy2(os.path.join(src_dir, 'versions.json'), os.path.join(dest_dir, 'versions.json'))

        self._encode_html_data(src_dir, 'help', dest_dir, header_field='help', file_number=self.help_file_number)
        self._encode_images(src_dir, dest_dir)
        self.encode_code(src_dir, _file_name)
        self._encode_info(src_dir, _file_name, dest_dir)
        self.write_encode_code(dest_dir)
        helper.brace_file_write(self.header['data'], dest_dir, self.header["file_uuid"])
        file_list.append(self.header["file_uuid"])

        file_list.append('versions')
        file_list.extend(self.file_list)
        helper.brace_file_write(self.encode_versions(file_list), dest_dir, 'versions')
        return None

    def encode_version(self):
        return self.header['version']

    def _decode_images(self, src_dir, dest_dir):
        if self._images:
            for elem in self._images:
                try:
                    data = helper.brace_file_read(src_dir, f'{self.header["uuid"]}.{self._images[elem]}')
                except FileNotFoundError:
                    return
                try:
                    if data[0][2] and data[0][2][0] and data[0][2][0][0]:
                        bin_data = self._extract_b64_data(data[0][2][0])
                        helper.bin_write(bin_data, dest_dir, f'{elem}')
                except IndexError:
                    pass
                self.header[f'image_{elem}'] = data

    def _encode_images(self, src_dir, dest_dir):
        if self._images:
            for elem in self._images:
                try:
                    bin_data = helper.bin_read(src_dir, f'{elem}')
                except FileNotFoundError:
                    bin_data = None
                header = self.header.get(f'image_{elem}')
                if header and len(header[0]) > 1 and bin_data:
                    header[0][2][0][0] += b64encode(bin_data).decode(encoding='utf-8')
                if header:
                    file_name = f'{self.header["uuid"]}.{self._images[elem]}'
                    helper.brace_file_write(header, dest_dir, file_name)
                    self.file_list.append(file_name)

    def fill_header_includes(self, include_index):
        header_data = self.header['data']
        index_includes_group = 2
        count_includes_group = int(header_data[0][index_includes_group])
        for index_group in range(count_includes_group):
            group = header_data[0][index_includes_group + index_group + 1]
            group_uuid = group[0]
            group_version = group[1][0]
            try:
                metadata_group = MetaDataGroup(group_uuid)
            except ValueError:
                raise ExtException(message='Неизвестная группа метаданных', detail=group_uuid)
            include = group[1][1] if group_version == '6' else group[1]
            try:
                count_include_types = int(include[2])
            except IndexError:
                raise ExtException(message='Include types not found', detail=self.__class__.__name__)
            for i in range(count_include_types):
                _metadata = include[i + 3]
                if isinstance(_metadata, str):
                    metadata_type = MetaDataTypes[_metadata]
                    include_objects = include_index.get(_metadata, [])
                    include[i + 3] = [metadata_type.value, str(len(include_objects)), *include_objects]



from .FormElement import FormElement, calc_offset, check_count_element
from ....helper import FuckingBrackets
from ....ext_exception import ExtException


class Decoration(FormElement):
    @classmethod
    def get_name_node_offset(cls, raw_data):
        return calc_offset([(3, 1), (1, 1), (2, 0)], raw_data)

    # @classmethod
    # def decode(cls, form, path, raw_data):
    #     # _version = raw_data[0]
    #     result = super().decode(form, path, raw_data)
    #     return result

    # @classmethod
    # def decode_5(cls, form, raw_data):
    #
    #     try:
    #         size = check_count_element([
    #             (3, 1), (1, 1), (15, 1)
    #         ], raw_data)
    #     except TypeError:
    #         raise FuckingBrackets(detail=cls.__name__)
    #     except Exception as err:
    #         raise ExtException(parent=err)
    #     if size != 34:
    #         raise FuckingBrackets(detail=cls.__name__)
    #     result = super().decode(form, raw_data)
    #     return result
    #
    # @classmethod
    # def decode_11(cls, form, raw_data):
    #     try:
    #         size = check_count_element([
    #             (3, 1), (1, 1), (15, 1), (5, 1)
    #         ], raw_data)
    #     except TypeError:
    #         raise FuckingBrackets(detail=cls.__name__)
    #     except Exception as err:
    #         raise ExtException(parent=err)
    #     if size != 33:
    #         raise FuckingBrackets(detail=cls.__name__)
    #     result = super().decode(form, raw_data)
    #     return result
    #
    # @classmethod
    # def decode_12(cls, form, raw_data):
    #     try:
    #         size = check_count_element([
    #             (3, 1), (1, 1), (15, 1), (5, 1)
    #         ], raw_data)
    #     except TypeError:
    #         raise FuckingBrackets(detail=cls.__name__)
    #     except Exception as err:
    #         raise ExtException(parent=err)
    #     if size != 34:
    #         raise FuckingBrackets(detail=cls.__name__)
    #     result = super().decode(form, raw_data)
    #     return result
    #
    # @classmethod
    # def _decode(cls, form, raw_data):
    #     try:
    #         size = check_count_element([
    #             (3, 1), (1, 1), (15, 1), (5, 1)
    #         ], raw_data)
    #     except TypeError:
    #         raise FuckingBrackets(detail=cls.__name__)
    #     except Exception as err:
    #         raise ExtException(parent=err)
    #     if size != 34:
    #         raise FuckingBrackets(detail=cls.__name__)
    #     result = super().decode(form, raw_data)
    #     return result
